#include "main.h"
#include "YSwitchProc.h"

volatile uint8_t g_digitData[3];

#define g_brightness 2

volatile uint8_t g_swPortState;

volatile uint8_t g_swState;
volatile uint8_t g_encState;

volatile uint16_t g_tempoInterval;
volatile uint8_t g_tempoIntervalResidue;
volatile uint8_t  g_tempoBeat; 
volatile uint8_t  g_universalTimer[2];

const uint8_t g_fontTable[] PROGMEM = {
	0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110, 
	0b01101101, 0b01111101, 0b00100111, 0b01111111, 0b01101111
	
};

static uint8_t GetFont(const uint8_t n) {
	return pgm_read_byte(&(g_fontTable[n]));
}

static void HexToBCD(uint8_t hex, uint8_t bin3[]);

sEEPSettings g_eepSettings;

uint8_t eep_dummy __attribute__((section(".eeprom")));
uint8_t eep_area[sizeof(g_eepSettings)] __attribute__((section(".eeprom")));

static void LoadEEP(void){ 
	uint8_t i;
	uint8_t* p = (uint8_t*)(&g_eepSettings);
	for(i=0; i<sizeof(g_eepSettings); i++){
		uint8_t a = eeprom_read_byte(&(eep_area[i]));
		if(a != 0xFF) p[i] = a;
	}
}
static void SaveEEP(void){ 
	uint8_t i;
	uint8_t* p = (uint8_t*)(&g_eepSettings);
	for(i=0; i<sizeof(g_eepSettings); i++){
		eeprom_write_byte(&(eep_area[i]), p[i]);
	}
}

static uint8_t GetSwitchStateEx(void){
	PORTB |= 0b00000111; 
	PORTA |= 0b01111111; 
	DDRA  &= 0b10000000; 
	
	_delay_us(10);
	uint8_t ret = PINA;

	DDRA  |= 0b01111111;
	return ret;
}

static void Timer1_OVF_vect_virtual(void){

	{
		volatile static uint16_t intervalCounter;
		static uint8_t residueCounter;
		
		intervalCounter ++;
		if(intervalCounter > g_tempoInterval){ 
			intervalCounter = 0;

			residueCounter += (64)-1;
			
			if(residueCounter >= g_tempoIntervalResidue){
				intervalCounter ++; 
			}

			g_tempoBeat = (g_tempoBeat+1) % g_eepSettings.curTime;

			if(g_tempoBeat){
				g_universalTimer[0] = 255-10;
			}else{
				g_universalTimer[0] = 255-110 ;
			}
		}
	}
	
	if(g_universalTimer[0] < 255) g_universalTimer[0]++;
	if(g_universalTimer[1] < 255) g_universalTimer[1]++;

}

ISR(TIMER0_OVF0_vect ){
	static uint8_t place;
	volatile static uint8_t subCnt;
	subCnt = (subCnt+1) % 6;
	
	if(subCnt == 0){

		g_swPortState = GetSwitchStateEx();
		
		static uint8_t subCnt_local;
		
		if(subCnt_local == 0){ 
			GetSwitchState(g_swPortState, (uint8_t*) &g_swState);
		}
		GetREncState(g_swPortState, (uint8_t*) &g_encState);
		subCnt_local = (subCnt_local+1) % 40;
		
		Timer1_OVF_vect_virtual();
	}else if(subCnt == 1){
		
		place += 1;
		if(place == 3) place = 0;

		PORTA = (PORTA & 0x80) | ~( g_digitData[place] );
	}
	
	if(subCnt > g_brightness){
		
		CBI(PORTB, 2-place);
	}

}

ISR(TIMER1_OVF1_vect ){
	
	if(OCR1A != 0) CBI(PORTA, PA7);
}

ISR(TIMER1_CMPA_vect ){
	
	SBI(PORTA, PA7);
}

static void HexToBCD(uint8_t hex, uint8_t bin3[]){

	bin3[0] = hex/100;
	bin3[1] = (hex/10)%10;
	bin3[2] = hex%10;
}
static void UpdateDisplay(uint8_t number){
	uint8_t buf[3];
	HexToBCD(number, buf);
	g_digitData[0] = buf[0]? GetFont(buf[0]): 0;
	g_digitData[1] = (buf[0]||buf[1])? GetFont(buf[1]): 0;
	g_digitData[2] = GetFont(buf[2]);
}
static void NotifyBell(void){
	OCR1A = 128; 
	OCR1B = 128;

	while(g_swState == 0){
		PWMEnable(); 
		_delay_ms(50);
		PWMDisable(); 
		_delay_ms(50);
	}
	
	g_swState = 0;
	
	OCR1A = 0; 
}

static void UpdateBPM(void){
	
	if(g_eepSettings.curTempo < TEMPO_RANGE_MIN) g_eepSettings.curTempo = TEMPO_RANGE_MIN;
	if(g_eepSettings.curTempo > TEMPO_RANGE_MAX) g_eepSettings.curTempo = TEMPO_RANGE_MAX;
	UpdateDisplay(g_eepSettings.curTempo);
	
	cli(); 
	
#ifdef ACTUAL_CLOCK_FREQ
	uint32_t temp = ((uint32_t)(120L*500L*256L*4/10 * ((double)ACTUAL_CLOCK_FREQ/4915200.0) ))
		/g_eepSettings.curTempo; 
#else
	uint32_t temp = ((uint32_t)(120L*500L*256L*4/10))
		/g_eepSettings.curTempo; 
#endif
	g_tempoInterval = temp >> 8;
	g_tempoIntervalResidue = temp;

	sei();
}

static void InitPort(void){

	DDRA  = 0b11111111;

	DDRB  = 0b01001111;
	PORTB = 0b01000111; 
	
}

static void Init(void){

	InitPort();

	uint8_t muteMode = (GetSwitchStateEx() & 0b00000100);
	
	while((GetSwitchStateEx() & 0b00001100) != 0b00001100);
	
	if(!muteMode) CBI(DDRB, PB3);

	TCCR0 = 0b00000010;

	TCCR1A = 0b00000001;
	TCCR1B = 0b00000100;
	OCR1C  = 153;

	TIMSK  = 0b01000110; 
	
	LoadEEP();

	if(g_eepSettings.curTime > 6) g_eepSettings.curTime = 4;
	
	sei();

}

int main(void){
	
	Init();
	
	uint16_t tempoValue_int = ((uint16_t) g_eepSettings.curTempo) << 2;
	uint8_t tempoRestoreValue; 
	
	g_universalTimer[0] = 255;

	uint8_t  g_countMode = 0;

	g_eepSettings.curMode  = MODE_METRONOME;
	g_eepSettings.timerSec = 0;
	
	while(1){
		
		if(g_eepSettings.curMode == MODE_METRONOME)
		{
			
			if(g_swState){
				uint8_t swState = g_swState;
				g_swState = SW_UNPUSHED;

				if(swState == (SW_SW2 | SW_LONG))
				{ 
					cli();
					SaveEEP();
					CBI(PORTB, PB6);
					while(1);
				}
				else if(swState == SW_SW2)
				{ 
					g_eepSettings.curMode = MODE_TIMER;
					tempoValue_int = ((uint16_t) g_eepSettings.timerMin) << 2;
					tempoRestoreValue = g_eepSettings.curTempo;
					g_eepSettings.curTempo = 60;
				}
				else if(swState == (SW_SW1 | SW_LONG))
				{ 
					g_tempoInterval = 1;
					_delay_ms(10);
					g_tempoBeat = 0;
				}
				else if(swState == SW_SW1)
				{ 
					g_eepSettings.curTime ++;
					if(g_eepSettings.curTime > 6) g_eepSettings.curTime = 3;

					g_universalTimer[1] = 255-100;
					UpdateDisplay(g_eepSettings.curTime);
					while(g_universalTimer[1] != 255);
					
				}
				
			}
			if(g_encState){
				uint8_t encState = g_encState;
				g_encState = RENC_NORM;

				if(encState == RENC_FWD)
				{
					tempoValue_int += 1;
				}
				else if(encState == RENC_REV)
				{
					tempoValue_int -= 1;
				}
				if(tempoValue_int > TEMPO_RANGE_MAX<<2) tempoValue_int = TEMPO_RANGE_MAX<<2;
				if(tempoValue_int < TEMPO_RANGE_MIN<<2) tempoValue_int = TEMPO_RANGE_MIN<<2;
				g_eepSettings.curTempo = tempoValue_int >> 2;
				
			}

			if(g_universalTimer[0] == 255) {
				OCR1A = 0;
				
				if(g_tempoBeat){
					PWMDisable();
					CBI(PORTB, PB3);
				}
			}
			else {
				uint16_t temp = (uint8_t)( (255-g_universalTimer[0])<<0);
				
				PWMEnable();
				if(g_tempoBeat == 0){
					OCR1B  = (((temp*temp)>>8)*temp)>>8; 
					if(temp>20) temp-=20;
					OCR1A  = temp;
				}else{
					OCR1B = 10;
					OCR1A = 60;
				}
			}
			
			UpdateBPM();
		} 
		else
		{ 
			
			if(g_swState){
				uint8_t swState = g_swState;
				g_swState = SW_UNPUSHED;

				if(swState == (SW_SW2 | SW_LONG))
				{ 
					cli();
					CBI(PORTB, PB6);
					while(1);
				}
				else if(swState == SW_SW2)
				{ 
					g_eepSettings.curTempo = tempoRestoreValue;
					tempoValue_int = ((uint16_t) g_eepSettings.curTempo) << 2;
					g_eepSettings.curMode = MODE_METRONOME;
				}
				else if(swState == SW_SW1)
				{ 
					g_countMode ++;
				}
				
			}
			if((g_countMode == 0) && (g_encState)){
				uint8_t encState = g_encState;
				g_encState = RENC_NORM;

				if(encState == RENC_FWD)
				{
					tempoValue_int += 1;
				}
				else if(encState == RENC_REV)
				{
					tempoValue_int -= 1;
				}
				g_eepSettings.timerMin = tempoValue_int / 4;
			}

			g_eepSettings.timerMin ? UpdateDisplay(g_eepSettings.timerMin): UpdateDisplay(g_eepSettings.timerSec);

			if(g_countMode & 1){ 
				if(g_universalTimer[0] != 255) {
					if(g_eepSettings.timerSec){
						g_eepSettings.timerSec -= 1;
					}else{
						if(g_eepSettings.timerMin){
							g_eepSettings.timerMin -= 1;
							g_eepSettings.timerSec = 59;
						}else{
							NotifyBell();
							g_countMode = 0;
						}
					}
					OCR1A = 100;
					_delay_ms(100);
					OCR1A = 0;
					g_universalTimer[0] = 255;
				}
			}
			PWMDisable();
			OCR1A = 0;
			
		}
	}

	return 0;
}
